#include "Amisic.H"

#include "Simple_Chain.H"
#include "Simple_String.H"

using namespace AMISIC;

Amisic::Amisic():
  m_hardmodel(Unknown),
  m_softmodel(Unknown),
  p_hardbase(NULL),
  p_softbase(NULL),
  p_model(NULL),
  p_beam(NULL),
  p_isr(NULL),
  m_external(false) {}

Amisic::Amisic(MODEL::Model_Base *const model,
	       BEAM::Beam_Spectra_Handler *const beam,
	       PDF::ISR_Handler *const isr):
  m_hardmodel(Unknown),
  m_softmodel(Unknown),
  p_hardbase(NULL),
  p_softbase(NULL),
  p_model(model),
  p_beam(beam),
  p_isr(isr),
  m_external(true) {}

Amisic::~Amisic() 
{
  if (p_hardbase!=NULL) delete p_hardbase;
  if (p_softbase!=NULL) delete p_softbase;
}

bool Amisic::Initialize()
{
  if (!CheckInputPath()) return false;
  if (!CheckInputFile()) return false;
  ATOOLS::Data_Reader *reader = new ATOOLS::Data_Reader("=",";","!");
  reader->SetInputPath(InputPath());
  reader->SetInputFile(InputFile());
  std::vector<std::string> model;
  if (!reader->VectorFromFile(model,"HARD_MODEL_NAME",ATOOLS::noinputtag,reader->VHorizontal)) {
    model.push_back("Simple_Chain");
  }
  for (size_t i=1;i<model.size();++i) model[0]+=std::string(" ")+model[i];
  SelectHardModel(StringToModelID(model[0]));
  if (!reader->VectorFromFile(model,"SOFT_MODEL_NAME",ATOOLS::noinputtag,reader->VHorizontal)) {
    model.push_back("None");
  }
  for (size_t i=1;i<model.size();++i) model[0]+=std::string(" ")+model[i];
  SelectSoftModel(StringToModelID(model[0]));
  std::string file;
  if (!reader->ReadFromFile(file,"HARD_MODEL_FILE")) file=InputFile();
  p_hardbase->SetInputPath(InputPath());
  p_hardbase->SetInputFile(file);
  if (!reader->ReadFromFile(file,"SOFT_MODEL_FILE")) file=InputFile();
  p_softbase->SetInputPath(InputPath());
  p_softbase->SetInputFile(file);
  delete reader;
  bool success=true;
  success=success&&p_hardbase->Initialize();
  success=success&&p_softbase->Initialize();
  return success;
}

void Amisic::SameHardProcess(ATOOLS::Blob *blob)
{
  p_hardbase->CreateBlob(blob);
}

void Amisic::SameSoftProcess(ATOOLS::Blob *blob)
{
  p_softbase->CreateBlob(blob);
}

bool Amisic::VetoHardProcess(ATOOLS::Blob *blob)
{
  return p_hardbase->VetoProcess(blob);
}

bool Amisic::VetoSoftProcess(ATOOLS::Blob *blob)
{
  return p_softbase->VetoProcess(blob);
}

bool Amisic::GenerateHardProcess(ATOOLS::Blob *blob)
{
  if (MI_Base::StopGeneration(MI_Base::HardEvent)) return false;
  if (!p_hardbase->DiceProcess()) return false;
  p_hardbase->UpdateAll(p_hardbase);
  return p_hardbase->CreateBlob(blob);
}

bool Amisic::GenerateSoftProcess(ATOOLS::Blob *blob)
{
  if (MI_Base::StopGeneration(MI_Base::SoftEvent)) return false;
  if (!p_softbase->DiceProcess()) return false;
  p_softbase->UpdateAll(p_softbase);
  return p_softbase->CreateBlob(blob);
}

bool Amisic::GenerateHardEvent(ATOOLS::Blob_List *blobs)
{
  p_hardbase->Reset();
  while (true) {
    ATOOLS::Blob *newblob = new ATOOLS::Blob();
    if (GenerateHardProcess(newblob)) {
      newblob->SetType(ATOOLS::btp::Hard_Collision);
      newblob->SetStatus(1);
      newblob->SetId();
      blobs->push_back(newblob);
    }
    else {
      delete newblob;
      if (MI_Base::StopGeneration(MI_Base::HardEvent)) return true;
      msg_Tracking()<<"Amisic::GenerateHardEvent(): "
		    <<"Cannot create hard underlying event."<<std::endl
		    <<"   Abort attempt."<<std::endl;
      return false;
    }
  }
  return true;
}

bool Amisic::GenerateSoftEvent(ATOOLS::Blob_List *blobs)
{
  p_softbase->Reset();
  while (true) {
    ATOOLS::Blob *newblob = new ATOOLS::Blob();
    if (GenerateSoftProcess(newblob)) {
      newblob->SetType(ATOOLS::btp::Soft_Collision);
      newblob->SetStatus(1);
      newblob->SetId();
      blobs->push_back(newblob);
    }
    else {
      delete newblob;
      if (MI_Base::StopGeneration(MI_Base::SoftEvent)) return true;
      msg_Tracking()<<"Amisic::GenerateSoftEvent(): "
		    <<"Cannot create soft underlying event."<<std::endl
		    <<"   Abort attempt."<<std::endl;
      return false;
    }
  } 
  return true;
}

bool Amisic::GenerateEvent(ATOOLS::Blob_List *blobs)
{
  Reset();
  if (!GenerateHardEvent(blobs)) return false;
  if (!GenerateSoftEvent(blobs)) return false;
  return true;
}

void Amisic::SameHardEvent(ATOOLS::Blob_List *blobs)
{
}

void Amisic::SameSoftEvent(ATOOLS::Blob_List *blobs)
{
}

void Amisic::Reset()
{
  MI_Base::ResetAll();
}

void Amisic::CleanUp()
{
  MI_Base::CleanUp();
}

bool Amisic::SelectHardModel(ModelID _m_hardmodel)
{ 
  m_hardmodel=_m_hardmodel; 
  if (p_hardbase!=NULL) delete p_hardbase;
  switch (m_hardmodel) {
  case SimpleChain:
    msg_Tracking()<<"Amisic::SelectHardModel("<<_m_hardmodel<<"): "
			  <<"Initialize simple hard underlying event model."<<std::endl;
    if (m_external) p_hardbase = new Simple_Chain(p_model,p_beam,p_isr);
    else p_hardbase = new Simple_Chain();
    break;
  case None:
    msg_Tracking()<<"Amisic::SelectHardModel("<<_m_hardmodel<<"): "
			  <<"Initialize no hard underlying event handler."<<std::endl;
    p_hardbase = new MI_None(MI_Base::HardEvent);
    break;
  default:
    ATOOLS::msg.Error()<<"Amisic::SelectHardModel("<<m_hardmodel<<"): "
		       <<"Model type inconsistent!"<<std::endl
		       <<"   Initialize no hard underlying event handler."<<std::endl;
    m_hardmodel=None;
    p_hardbase = new MI_None(MI_Base::HardEvent);
    break;
  }
  p_hardbase->SetInputPath(InputPath());
  p_hardbase->SetOutputPath(OutputPath());
  p_hardbase->SetInputFile(InputFile());
  return true;
}

bool Amisic::SelectSoftModel(ModelID _m_softmodel)
{ 
  m_softmodel=_m_softmodel; 
  if (p_softbase!=NULL) delete p_softbase;
  switch (m_softmodel) {
  case SimpleString:
    msg_Tracking()<<"Amisic::SelectSoftModel("<<_m_softmodel<<"): "
			  <<"Initialize simple soft underlying event model."<<std::endl;
    if (m_external) p_softbase = new Simple_String();
    else p_softbase = new Simple_String();
    break;
  case None:
    msg_Tracking()<<"Amisic::SelectSoftModel("<<_m_softmodel<<"): "
			  <<"Initialize no soft underlying event handler."<<std::endl;
    p_softbase = new MI_None(MI_Base::SoftEvent);
    break;
  default:
    ATOOLS::msg.Error()<<"Amisic::SelectSoftModel("<<m_softmodel<<"): "
		       <<"Model type inconsistent!"<<std::endl
		       <<"   Initialize no soft underlying event handler."<<std::endl;
    m_softmodel=None;
    p_softbase = new MI_None(MI_Base::SoftEvent);
    break;
  }
  p_softbase->SetInputPath(InputPath());
  p_softbase->SetOutputPath(OutputPath());
  p_softbase->SetInputFile(InputFile());
  return true;
}

Amisic::ModelID Amisic::StringToModelID(std::string model) 
{
  if (model=="Simple_Chain") return SimpleChain;
  if (model=="Simple_String") return SimpleString;
  if (model=="None")         return None;
  ATOOLS::msg.Error()<<"Amisic::StringToModelID("<<model<<"): "
		     <<"Model type unknown!"<<std::endl
		     <<"   Returning 'Unknown'."<<std::endl;
  return Unknown;
}

std::string Amisic::ModelIDToString(ModelID model) 
{
  switch (model) {
  case SimpleChain : return std::string("Simple_Chain");
  case SimpleString: return std::string("Simple_String");
  case None        : return std::string("None");
  case Unknown     : return std::string("Unknown");
  }
  return std::string("Unknown");
}
